
/* compile with
 * $ cc -DSSD_IO_MACROS=\"test-ssd.h\" -std=c99 test-ssd.c ssd1963.c -o test-ssd
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>

#include "ssd1963.h"

/* HSD050IDW1-A 5.0" TFT LCD */
#define HSD050IDW1_A \
	(struct ssd_display){ \
		SSD_LCD_24BIT | SSD_LCD_MODE_TFT, \
		{ 800, 40, 48, 40 }, \
		{ 480, 13,  3, 29 }, \
		0, 33e6, 40e6, \
	}

/* HSD043I9W1-A 4.3" TFT LCD */
/* the datasheet doesn't specify typical values for hold times of the sync
 * signals and includes these times in the back porch times;
 * minimum is specified to 1 for hsync as well as vsync */
#define HSD043I9W1_A \
	(struct ssd_display){ \
		SSD_LCD_24BIT | SSD_LCD_MODE_TFT, \
		{ 480, 5, 8, 40 - 8 }, \
		{ 272, 8, 1,  8 - 1 }, \
		5e6, 9e6, 12e6, \
	}

#define ITDB02_XTAL_FREQ	12000000

int main(int argc, char **argv)
{
	struct ssd_init_vector iv;
	enum ssd_err r;
	uint_least32_t pclk;
	uint_least64_t pclk_scaled;

	if (argc < 3 || argc > 4) {
		fprintf(stderr, "usage: %s <pll_m> <pll_n> [<refresh_rate>]\n",
			argv[0]);
		return EXIT_FAILURE;
	}

	memset(&iv, 0, sizeof(iv));
#if 1
	r = ssd_iv_init(&iv,
		ITDB02_XTAL_FREQ,
		atoi(argv[1]), atoi(argv[2]), 1,
		&HSD050IDW1_A, argc > 3 ? atoi(argv[3]) : 0);
	if (r != SSD_ERR_NONE) {
		fprintf(stderr, "error %d init'ing iv: %s\n", r, ssd_strerr(r));
		if (r != SSD_ERR_PXCLK_OOR)
			return EXIT_FAILURE;
	}
#else
	iv.pll_m         = atoi(argv[1]);
	iv.pll_n         = atoi(argv[2]);
	iv.pll_as_sysclk = 1;
	iv.in_clk_freq   = ITDB02_XTAL_FREQ;

	ssd_iv_set_display(&iv, &HSD050IDW1_A);

	pclk             = ssd_iv_calc_pixel_freq(&iv,
		argc > 3 ? atoi(argv[3]) : 0);
	iv.lshift_mult   = ssd_iv_calc_lshift_mult(&iv, pclk);
#endif
	pclk_scaled      = ssd_iv_get_pixel_freq_frac(&iv);
	fprintf(stderr,
		"in_clk_freq: %g MHz, "
		"PLL: %" PRIdLEAST32 "/%" PRIdLEAST32 " -> %g MHz, "
		"VCO: %g MHz\n"
		"sys: %g MHz, "
		"px clk: req'ed: %" PRIdLEAST32 ", set: %" PRIdLEAST64 ".%" PRIdLEAST64 " Hz, "
		"lshift: %" PRIdLEAST32 ", rate: %g Hz\n"
		"ht, hps, hpw, lps, lpspp: %" PRIdLEAST16 " %" PRIdLEAST16 " %" PRIdLEAST16 " %" PRIdLEAST16 " %" PRIdLEAST16 "\n"
		"vt, vps, vpw, fps       : %" PRIdLEAST16 " %" PRIdLEAST16 " %" PRIdLEAST16 " %" PRIdLEAST16 "\n"
		"hdp: %" PRIdLEAST16 ", "
		"vdp: %" PRIdLEAST16 "\n",
		iv.in_clk_freq / 1e6,
		iv.pll_m, iv.pll_n, ssd_iv_get_pll_freq(&iv) / 1e6,
		ssd_iv_get_vco_freq(&iv) / 1e6,
		ssd_iv_get_sys_freq(&iv) / 1e6,
		pclk, pclk_scaled >> 20, pclk_scaled & ((1L << 20) - 1),
		iv.lshift_mult, pclk_scaled / (0x1p20 * iv.ht * iv.vt),
		iv.ht, iv.hps, iv.hpw, iv.lps, iv.lpspp,
		iv.vt, iv.vps, iv.vpw, iv.fps,
		iv.hdp,
		iv.vdp);

	if (r != SSD_ERR_NONE)
		return EXIT_FAILURE;

	r = ssd_init_pll(&iv);
	if (r != SSD_ERR_NONE) {
		fprintf(stderr, "error %d init'ing PLL: %s\n",
			r, ssd_strerr(r));
		return EXIT_FAILURE;
	}

	r = ssd_init_display(&iv);
	if (r != SSD_ERR_NONE) {
		fprintf(stderr, "error %d init'ing display: %s\n",
			r, ssd_strerr(r));
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
